#include "c4d_baseobject.h"
#include "c4d_basedocument.h"
#include "c4d_general.h"
#include "c4d_objectdata.h"
#include "newobj.h"

#include "maxon/errorbase.h"
#include "main.h"
//#include "ooptimizeobject.h"

class OptimizeObject : public ObjectData
{
public:

  virtual Bool Init(GeListNode* node, Bool isCloneInit) { return true; }
  virtual BaseObject* GetVirtualObjects(BaseObject* op, const HierarchyHelp* hh);

  static NodeData* Alloc() { return NewObjClear(OptimizeObject); }
};


BaseObject* OptimizeObject::GetVirtualObjects(BaseObject* op, const HierarchyHelp* hh)
{
  // Declare our object and document.
  PolygonObject* obj = nullptr;
  BaseDocument* temp = nullptr;

  // Scope handlers should be at the top of a scope.
  iferr_scope_handler
  {
    DiagnosticOutput("@: @", MAXON_FUNCTIONNAME, err);
    if (obj)
      PolygonObject::Free(obj);
    if (temp)
      BaseDocument::Free(temp);
    return BaseObject::Alloc(Onull);
  };

  // Create the polygon object and check for failures along the way. You could of course 'optimize'
  // these out, but then your plugin will crash when things go south (machine runs out of memory
  // for example).
  obj = PolygonObject::Alloc(8, 2);
  if (!obj)
    iferr_throw(maxon::OutOfMemoryError(
      MAXON_SOURCE_LOCATION, "Could not allocate polygon object."_s));

  if (!obj->ResizeObject(8, 2))
    iferr_throw(maxon::OutOfMemoryError(
      MAXON_SOURCE_LOCATION, "Could not resize polygon object."_s));

  CPolygon* const polygons = obj->GetPolygonW();
  Vector* const points = obj->GetPointW();
  if (!polygons || !points)
    iferr_throw(maxon::UnexpectedError(
      MAXON_SOURCE_LOCATION, "Could not access polygon data."_s));

  polygons[0].a = 0;
  polygons[0].b = 1;
  polygons[0].c = 2;
  polygons[0].d = 3;
  polygons[1].a = 4;
  polygons[1].b = 5;
  polygons[1].c = 6;
  polygons[1].d = 7;
  points[0] = Vector(0, 0, 0);
  points[1] = Vector(100, 0, 0);
  points[2] = Vector(100, 100, 0);
  points[3] = Vector(0, 100, 0);
  points[4] = Vector(0, 0, 0);
  points[5] = Vector(-100, 0, 0);
  points[6] = Vector(-100, 100, 0);
  points[7] = Vector(0, 100, 0);
 
  // Allocate the dummy document and insert the object.
  temp = BaseDocument::Alloc();
  if (!temp)
    iferr_throw(maxon::OutOfMemoryError(
      MAXON_SOURCE_LOCATION, "Could not allocate dummy document."_s));

  temp->InsertObject(obj, nullptr, nullptr);

  // Run 'Optimize' on these inputs.
  ModelingCommandData mcdOptimize;
  mcdOptimize.doc = temp;
  mcdOptimize.op = obj;
  if (!SendModelingCommand(MCOMMAND_OPTIMIZE, mcdOptimize))
    iferr_throw(maxon::UnexpectedError(
      MAXON_SOURCE_LOCATION, "Failed to run 'Optimize' command."_s));

  // I am seeing this just now, but running Explode Segments makes zero sense for polygons for me, as this
  // is a spline tool. Am I overlooking here something?
  ModelingCommandData mcdExplodeSegments;
  mcdExplodeSegments.doc = temp;
  mcdExplodeSegments.op = obj;
  if (!SendModelingCommand(MCOMMAND_EXPLODESEGMENTS, mcdExplodeSegments))
    iferr_throw(maxon::UnexpectedError(
      MAXON_SOURCE_LOCATION, "Failed to run 'Explode Segments' command."_s));

  // Remove the object and free the document, as everything went fine.
  obj->Remove();
  BaseDocument::Free(temp);

  return obj;
}

// Be sure to use a unique ID obtained from www.plugincafe.com
static const Int32 g_pid_optimize_object = 1061894;
Bool RegisterOptimizeObject()
{
  return RegisterObjectPlugin(
    g_pid_optimize_object, "Optimize Object"_s, OBJECT_GENERATOR, OptimizeObject::Alloc,
    "ooptimizeobject"_s, nullptr, 0);
}
